
/*
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1998-2000 by The Regents of the
 *  University of California. All rights reserved. See the COPYRIGHT
 *  file for details.
 *
 */

#include "base/polygraph.h"

#include "xstd/h/math.h"
#include "xstd/h/iomanip.h"

#include "xstd/Clock.h"
#include "base/Progress.h"
#include "base/polyLogCats.h"
#include "base/polyLogTags.h"
#include "pgl/StatsSampleCfg.h"
#include "pgl/StatsSampleSym.h"
#include "pgl/GoalSym.h"
#include "pgl/PhaseSym.h"
#include "runtime/Connection.h"
#include "runtime/Goal.h"
#include "runtime/httpHdrs.h"
#include "runtime/Xaction.h"
#include "runtime/IcpXaction.h"
#include "runtime/LogComment.h"
#include "runtime/SharedOpts.h"
#include "runtime/PolyOLog.h"
#include "runtime/Rptmstat.h"
#include "runtime/StatsSampleMgr.h"
#include "runtime/StatPhaseMgr.h"
#include "runtime/StatPhase.h"
#include "runtime/globals.h"
#include "runtime/polyErrors.h"
#include "runtime/polyBcastChannels.h"

static enum { wssIgnore, wssWaitFreeze, wssFrozen } TheWssState = wssIgnore;

StatPhase::StatPhase(): theRecs(lgcEnd), theMgr(0), theRptmstat(0),
	theLoadFactor(this), theRecurFactor(this), theSpecialMsgFactor(this),
	wasUsed(false), logStats(true), waitWssFreeze(false), 
	readyToStop(false), wasStopped(false), theLockCount(0) {

	for (int i = 0; i < theRecs.capacity(); ++i)
		theRecs.push(new StatPhaseRec);

	theChannels.append(TheInfoChannel);
}

StatPhase::~StatPhase() {
	delete theRptmstat; theRptmstat = 0;
	while (theRecs.count()) delete theRecs.pop();
}

void StatPhase::configure(const PhaseSym *cfg, const StatPhase *prevPh) {
	Assert(cfg);

	name(cfg->name());
	theGoal.configure(cfg->goal());
	for (int i = 0; i < theRecs.count(); ++i)
		theRecs[i]->theGoal = theGoal;

	cfg->logStats(logStats);
	cfg->waitWssFreeze(waitWssFreeze);

	double b = -1;
	double e = -1;
	const bool varyLoad = ((int)cfg->loadFactorBeg(b) + (int)cfg->loadFactorEnd(e)) != 0;
	theLoadFactor.configure(b, e, prevPh ? &prevPh->theLoadFactor : 0);
	checkFactor(theLoadFactor, varyLoad, "load");

	b = -1;
	e = -1;
	const bool varyRecur = ((int)cfg->recurFactorBeg(b) + (int)cfg->recurFactorEnd(e)) != 0;
	theRecurFactor.configure(b, e, prevPh ? &prevPh->theRecurFactor : 0);
	checkFactor(theRecurFactor, varyRecur, "recurrence");

	b = -1;
	e = -1;
	const bool varySpReq = ((int)cfg->specialMsgFactorBeg(b) + (int)cfg->specialMsgFactorEnd(e)) != 0;
	theSpecialMsgFactor.configure(b, e, prevPh ? &prevPh->theSpecialMsgFactor : 0);
	checkFactor(theSpecialMsgFactor, varySpReq, "special_req");

	if (RptmstatSym *s = cfg->rptmstat()) {
		theRptmstat = new Rptmstat;
		theRptmstat->configure(*s);
	}

	configureSamples(cfg);
}

void StatPhase::configureSamples(const PhaseSym *cfg) {
	Array<const StatsSampleSym*> ssyms;
	if (cfg->statsSamples(ssyms)) {
		for (int i = 0; i < ssyms.count(); ++i) {
			const StatsSampleSym &ssym = *ssyms[i];

			StatsSampleCfg cfg;
			// XXX: append counter to phase names
			cfg.name = ssym.name() ? ssym.name() : name();
			Assert(ssym.capacity(cfg.capacity));
			cfg.start = ssym.start();

			if (cfg.start >= 0) {
				TheStatsSampleMgr.addSample(cfg);
			} else {
				theSamples.append(new StatsSampleCfg(cfg));
				TheStatsSampleMgr.willAddSample();
			}
		}
	}
	// no support for multiple pending samples for now
	Assert(theSamples.count() <= 1); 
}

void StatPhase::checkFactor(TransFactor &, bool varyF, const String &label) const {
	// varying load is possible only if goal is specified
	if (!theGoal && varyF) {
		cerr << "phase named `" << theName
			<< "' has " << label
			<< " factors but has no positive goal" << endl;
		exit(-2);
	}
}

void StatPhase::lock() {
	theLockCount++;
}

void StatPhase::unlock() {
	if (locked()) {
		theLockCount--;
		if (!locked())
			checkpoint();
	} else {
		Comment(0) << "internal error: attempt to unlock an unlocked phase (salvaged)" << endc;
	}
}

bool StatPhase::locked() const {
	return theLockCount > 0;
}

Time StatPhase::duration() const {
	return TheClock.time() >= theIntvlStart ?
		TheClock - theIntvlStart : Time(0,0);
}

int StatPhase::xactCnt() const {
	int count = 0;
	for (int i = 0; i < theRecs.count(); ++i)
		count += theRecs[i]->xactCnt();
	return count;
}

BigSize StatPhase::fillSz() const {
	BigSize fill(0);
	for (int i = 0; i < theRecs.count(); ++i)
		fill += theRecs[i]->totFillSize();
	return fill;
}

int StatPhase::fillCnt() const {
	int fill = 0;
	for (int i = 0; i < theRecs.count(); ++i)
		fill += theRecs[i]->totFillCount();
	return fill;
}

int StatPhase::errCnt() const {
	int count = 0;
	for (int i = 0; i < theRecs.count(); ++i)
		count += theRecs[i]->theXactErrCnt;
	return count;
}

double StatPhase::errRatio() const {
	int xcount = 0, ecount = 0;
	for (int i = 0; i < theRecs.count(); ++i) {
		xcount += theRecs[i]->xactCnt();
		ecount += theRecs[i]->theXactErrCnt;
	}
	return Ratio(ecount, xcount);
}


void StatPhase::start(StatPhaseMgr *aMgr, const StatPhase *prevPhase) {
	Assert(!theMgr);
	theMgr = aMgr;

	if (prevPhase) {
		for (int i = 0; i < theRecs.count(); ++i)
			theRecs[i]->keepLevels(*prevPhase->theRecs[i]);
	}

	// XXX: will not lock if fill phase finishes before we get ieWssFill
	if (waitWssFreeze && TheWssState == wssWaitFreeze)
		lock();

	if (theRptmstat)
		theRptmstat->start(this);

	TheOLog << bege(lgStatPhaseBeg, lgcAll) << name() << ende;

	startListen();
	theIntvlStart = TheClock;

	for (int i = 0; i < theSamples.count(); ++i) {
		StatsSampleCfg &cfg = *theSamples[i];
		if (cfg.start < 0) {
			cfg.start = theIntvlStart;
			if (theGoal.duration() > 0)
				cfg.start += theGoal.duration()/2; // middle of the phase
			cfg.start -= Clock::TheStartTime;
		}
		TheStatsSampleMgr.addSample(cfg);
	}

	// manager should be careful with immediate "returns"
	if (!checkpoint()) { 
		// not an immediate exit
		wasUsed = true; 
		if (theGoal.duration() > 0)
			sleepFor(theGoal.duration());
	}
}

void StatPhase::noteConnEvent(BcastChannel *ch, const Connection *c) {
	StatIntvl::noteConnEvent(ch, c);
	if (ch == TheConnCloseChannel && !c->bad()) {
		StatPhaseRec &rec = *theRecs[c->logCat()];
		const Time ttl = TheClock - c->openTime();
		rec.theConnClose.record(c->closeKind(), ttl, c->useCnt());
	}
	checkpoint();
}

void StatPhase::noteXactEvent(BcastChannel *ch, const Xaction *x) {
	StatIntvl::noteXactEvent(ch, x);

	if (ch == TheXactEndChannel) {
		StatPhaseRec &rec = *theRecs[x->logCat()];
		const ObjId &oid = x->oid();

		// stats must be recorded in only one category for totals to work!
		if (oid.basic()) {
			rec.theBasicXacts.record(x->lifeTime(), x->repSize(), oid.hit());
			rec.theContType.record(oid.type(), x->repSize());
		} else
		if (oid.repToRedir())
			rec.theRepToRedirXacts.record(x->lifeTime(), x->repSize());
		else
		if (oid.rediredReq())
			rec.theRediredReqXacts.record(x->lifeTime(), x->repSize());
		else
		if (oid.ims())
			rec.theImsXacts.record(x->lifeTime(), x->repSize(), x->httpStatus() == RepHdr::sc200_OK);
		else
		if (oid.reload())
			rec.theReloadXacts.record(x->lifeTime(), x->repSize());
		else {
			Assert(false); // all categories should be accounted for
		}
	}

	checkpoint();
}

void StatPhase::noteIcpXactEvent(BcastChannel *ch, const IcpXaction *x) {
	StatIntvl::noteIcpXactEvent(ch, x);

	if (ch == TheIcpXactEndChannel) {
		StatPhaseRec &rec = *theRecs[x->logCat()];
		if (!x->timedout())
			rec.theIcpXacts.record(x->lifeTime(), x->repSize(), x->hit());
	} else {
		Assert(false);
	}
	checkpoint();
}

void StatPhase::noteInfoEvent(BcastChannel *ch, InfoEvent ev) {
	Assert(ch == TheInfoChannel);
	if (ev == ieWssFill) {
		if (waitWssFreeze)
			lock();
		TheWssState = wssWaitFreeze;
	} else
	if (ev == ieWssFreeze) {
		if (waitWssFreeze)
			unlock();
		TheWssState = wssFrozen;
	}
}

void StatPhase::name(const String &aName) {
	theName = aName;
	if (!theName)
		theName = "";
	for (int i = 0; i < theRecs.count(); ++i)
		theRecs[i]->theName = theName;
}

OidGenStat &StatPhase::oidGenStat() {
	return theRecs[lgcCltSide]->theOidGen;
}

ErrorStat &StatPhase::errors(int logCat) {
	return theRecs[logCat]->theErrors;
}

void StatPhase::statsLogged(bool are) {
	logStats = are;
}

void StatPhase::wakeUp(const Alarm &alarm) {
	StatIntvl::wakeUp(alarm);
	// checkpoint() may lock the phase
	Assert(locked() || checkpoint() || locked());
}

void StatPhase::flush() {
	stop();

	if (logStats)
		storeAll(TheOLog, lgStatPhaseRec);
}

// the caller should not use the phase if checkpoint returns true
bool StatPhase::checkpoint()  {
	if (!locked() && theGoal.reached(*this)) {

		if (TheOpts.doSyncPhases) {
			if (!readyToStop) {
				Comment(5) << "fyi: local phase `" << name()
					<< "' reached synchronization point" << endc;
				lock(); // somebody needs to unlock us to move on
				readyToStop = true;
				return false;
			}
		}

		stop();
		report();

		Assert(theMgr);
		StatPhaseMgr *mgr = theMgr;
		theMgr = 0;
		mgr->noteDone(this);
		return true;
	}
	return false;
}

void StatPhase::stop() {
	Assert(theIntvlStart >= 0);
	if (wasStopped) // already stopped
		return;
	wasStopped = true;

	if (theRptmstat)
		theRptmstat->stop(this);

	setDuration(theIntvlStart);
	stopListen();
	TheAlarmClock.cancelAll(this);

	TheOLog << bege(lgStatPhaseEnd, lgcAll) << name() << ende;
}

StatIntvlRec &StatPhase::getRec(int cat) {
	Assert(0 < cat && cat < theRecs.count());
	return *theRecs[cat];
}

const StatIntvlRec &StatPhase::getRec(int cat) const {
	Assert(0 < cat && cat < theRecs.count());
	return *theRecs[cat];
}

// XXX: we probably should report _all_ active sides
void StatPhase::report() const {
	Comment(1) << "p-" << name();
	Comment << ' ' << setw(6) << TheProgress.xacts();
	if (TheReportCat > 0)
		theRecs[TheReportCat]->linePrint(Comment);
	Comment << endc;
}


void StatPhase::reportCfg(ostream &os) const {
	os 
		<< setw(10) << (const char*)name()
		<< ' ' << setw(8) << theLoadFactor.beg()
		<< ' ' << setw(8) << theLoadFactor.end()
		<< ' ' << setw(8) << theRecurFactor.beg()
		<< ' ' << setw(8) << theRecurFactor.end()
		<< ' ' << setw(8) << theSpecialMsgFactor.beg()
		<< ' ' << setw(8) << theSpecialMsgFactor.end()
		<< ' ' << setw(5) << theGoal
		<< "\t" 
			<< (logStats ? "" : " !log")
			<< (waitWssFreeze ? " wwss" : "")
		<< endl;
}
